#version 430 core

#define MAX_LIGHTS 10
#define SHADOW_MAP_CASCADES 5

in vec3 eyeDirection_cameraspace;
in vec3 normal_cameraspace;
in vec2 fragUV;
in vec3 fragPosition_cameraspace;
in vec3 fragPosition_worldspace;

uniform struct PointLight {
	vec3 position;
	vec3 diffColor;
	vec3 specColor;
	vec3 attenuation;
} pointLights[MAX_LIGHTS];

uniform bool bonfireLightEnabled;
uniform PointLight bonfireLight;

uniform struct DirectionalLight {
	vec3 direction;
	vec3 diffColor;
	vec3 specColor;
} directionalLights[MAX_LIGHTS];

uniform DirectionalLight shadowCasterDirectionalLight;

uniform uint nbPointLights;
uniform uint nbDirectionalLights;

uniform bool depthPass;

uniform bool nightMode;

uniform vec3 matAmbColor;
uniform vec3 matDiffColor;
uniform vec3 matSpecColor;
uniform float matShininess;
uniform vec3 matIlluminationWeights;
uniform float shadowMapCascades[SHADOW_MAP_CASCADES + 1];
uniform mat4 lightViewProj[SHADOW_MAP_CASCADES];
layout(binding = 0) uniform sampler2D tex;
layout(binding = 1) uniform sampler2DShadow shadowMapTex[SHADOW_MAP_CASCADES];

uniform mat4 M;
uniform mat4 V;

out vec4 color;


const vec2 poissonDisk[32] = vec2[]( 
   vec2(0.3568085f, 0.8572467f),
	vec2(0.5129315f, 0.3563072f),
	vec2(0.7601048f, 0.5944152f),
	vec2(0.1582207f, 0.6476642f),
	vec2(-0.1214974f, 0.9103214f),
	vec2(0.2900027f, 0.1111842f),
	vec2(0.7549803f, 0.0599489f),
	vec2(0.01185714f, 0.2484916f),
	vec2(-0.1212099f, 0.5640326f),
	vec2(-0.3450524f, 0.7177721f),
	vec2(0.8867018f, 0.3058905f),
	vec2(-0.3410425f, 0.1703072f),
	vec2(-0.5904795f, 0.3476484f),
	vec2(-0.09324571f, -0.1674143f),
	vec2(0.1812744f, -0.1945115f),
	vec2(-0.619292f, 0.7749628f),
	vec2(0.6588725f, -0.2548576f),
	vec2(0.9613882f, -0.169785f),
	vec2(0.6358203f, -0.7547918f),
	vec2(0.8861904f, -0.4308046f),
	vec2(0.3367413f, -0.4802357f),
	vec2(-0.3530865f, -0.5358725f),
	vec2(-0.4000572f, -0.2684504f),
	vec2(0.066328f, -0.5461764f),
	vec2(-0.6159387f, -0.08338374f),
	vec2(-0.9604711f, 0.07623308f),
	vec2(0.1423359f, -0.8516595f),
	vec2(-0.7081946f, -0.5865051f),
	vec2(-0.8030562f, -0.3231609f),
	vec2(-0.1948679f, -0.9543302f),
	vec2(-0.8458716f, 0.5325567f),
	vec2(-0.5556234f, -0.8252568f)
);

void computeDirectionalLightContribution(DirectionalLight light, vec3 n, out vec3 diffuse, out vec3 specular) {
	vec3 l = normalize(vec3(V * vec4(light.direction, 0))); //normalize because interpolated
	// cosine of the angle between the normal and the light direction (lambert)
	float cosTheta = clamp(dot(l, n), 0, 1);
	
	vec3 lightCol = light.diffColor;
	if(nightMode){
		lightCol = vec3(0.12, 0.18, 0.447);
	}	
	diffuse = lightCol * cosTheta;

	// calculate specular parts

	// Eye vector (towards the camera)
	vec3 E = normalize(eyeDirection_cameraspace);
	// Direction in which the triangle reflects the light
	vec3 R = reflect(-l, n);
	// Cosine of the angle between the Eye vector and the Reflect vector
	float cosAlpha = clamp( dot( E, R ), 0, 1 );

	specular = light.specColor * pow(cosAlpha, matShininess);
}

void computePointLightContribution(PointLight light, vec3 n, out vec3 diffuse, out vec3 specular) {
	vec3 l = (V * vec4(light.position, 1)).xyz - fragPosition_cameraspace; 
	float d = length(l);
	l = normalize(l);
	// cosine of the angle between the normal and the light direction (lambert)
	float cosTheta = clamp(dot(l,n), 0, 1);

	float totalAttenuation = light.attenuation[0] + light.attenuation[1] * d + light.attenuation[2] * d * d;

	diffuse = (1.0 / totalAttenuation) * light.diffColor * cosTheta;

	// calculate specular parts

	// Eye vector (towards the camera)
	vec3 E = normalize(eyeDirection_cameraspace);
	// Direction in which the triangle reflects the light
	vec3 R = reflect(-l, n);
	// Cosine of the angle between the Eye vector and the Reflect vector,
	float cosAlpha = clamp( dot( E, R ), 0,1 );

	specular = (1.0 / totalAttenuation) * light.specColor * pow(cosAlpha, matShininess);
}

float sampleSM(sampler2DShadow shadowMap, vec4 coords, float radius, int samples, float bias) {
	float visibility = 1.0f;
	// poisson sampling (with fixed pattern, no random seeding)
	float portion = (1.0 / float(samples));
	for (int i = 0; i < samples; i++) {
		visibility -= portion * (1 - textureProj(shadowMap, vec4(coords.xy + poissonDisk[i].xy * radius, coords.z, coords.w)));
		/*if(texture(shadowMap, coords.xy + poissonDisk[i].xy * radius).r < coords.z - bias) {
			visibility -= portion;
		}*/
	}
	return visibility;
}

void computeFogFactors(out float extinction, out float inscattering) {
	float z = length(fragPosition_cameraspace);

	float de = 0.025 * smoothstep(0.0, 6.0, 10.0 - fragPosition_worldspace.y);
	float di = 0.045 * smoothstep(0.0, 70.0, 10.0 - fragPosition_worldspace.y);

	extinction = exp(-z * de);
	inscattering = exp(-z * di);
}

void main() {
	if(depthPass) {
		return;
	}
		
	vec3 texture = texture(tex, fragUV).rgb;	
	//texture = vec3(1);

	vec3 n = normalize(normal_cameraspace); //normalize because interpolated
		
	vec3 diffuseIntensities = vec3(0);
	vec3 specularIntensities = vec3(0);
	vec3 diffuse;
	vec3 specular;
	for (uint i = 0; i < nbDirectionalLights; i++) {	
		computeDirectionalLightContribution(directionalLights[i], n, diffuse, specular);	
		diffuseIntensities += diffuse;
		specularIntensities += specularIntensities;
	}

	computeDirectionalLightContribution(shadowCasterDirectionalLight, n, diffuse, specular);	
	diffuseIntensities += diffuse;
	specularIntensities += specularIntensities;

	for (uint i = 0; i < nbPointLights; i++) {		
		computePointLightContribution(pointLights[i], n, diffuse, specular);
		diffuseIntensities += diffuse;
		specularIntensities += specularIntensities;
	}

	if(bonfireLightEnabled) {
		computePointLightContribution(bonfireLight, n, diffuse, specular);
		diffuseIntensities += diffuse;
		specularIntensities += specularIntensities;
	}

	// CSM
	bool showCascades = false;

	float visibility = 1;
	vec3 cascadeColor = vec3(1);
	for(int i = 0; i < SHADOW_MAP_CASCADES; i++) {
		if(fragPosition_cameraspace.z > shadowMapCascades[i + 1]) { // looking along the -z axis
			vec4 shadowMapCoord = lightViewProj[i] * vec4(fragPosition_worldspace, 1);
			visibility = sampleSM(shadowMapTex[i], shadowMapCoord, 0.0034, 32, 0);
			visibility = clamp(visibility, 0.45, 1.0);

			if(i == 0) 
				cascadeColor = vec3(0, 1, 0);
			else if(i == 1) 
				cascadeColor = vec3(1, 0, 0);
			else if(i == 2) 
				cascadeColor = vec3(0, 0, 1);
			else if(i == 3) 
				cascadeColor = vec3(1, 1, 0);

			break;
		}		
	}

	// final color
	color = vec4((
				texture * matIlluminationWeights.x * 0.3 + 
				texture * matDiffColor * matIlluminationWeights.y * diffuseIntensities * visibility +
				texture * matSpecColor * matIlluminationWeights.z * specularIntensities * visibility
			), 
		1);


	if(showCascades) {
		color.rgb = color.rgb * 0.8 + cascadeColor * 0.2;
	}
	
	// fog
	float fog_extinction = 1;
	float fog_inscattering = 1;

	computeFogFactors(fog_extinction, fog_inscattering);

	color.rgb = color.rgb * fog_extinction + vec3(0.7, 0.8, 0.9) * (1.0 - fog_inscattering);
}

