#version 430 core

#define MAX_LIGHTS 10

in vec3 normal;

in vec2 UV;

in vec3 eyeDirection_cameraspace;
in vec3 normal_cameraspace;
in vec3 fragPosition_cameraspace;

uniform struct PointLight {
	vec3 position;
	vec3 diffColor;
	vec3 specColor;
	vec3 attenuation;
} pointLights[MAX_LIGHTS];

uniform struct DirectionalLight {
	vec3 direction;
	vec3 diffColor;
	vec3 specColor;
} directionalLights[MAX_LIGHTS];

uniform struct SunLight {
	vec3 position;
	vec3 direction;
	vec3 diffColor;
	vec3 specColor;
	vec3 attenuation;
} sunLight;

uniform uint nbPointLights;
uniform uint nbDirectionalLights;

uniform vec3 matAmbColor;
uniform vec3 matDiffColor;
uniform vec3 matSpecColor;
uniform float matShininess;
uniform vec3 matIlluminationWeights;

uniform mat4 V;

out vec4 color;

uniform sampler2D surfaceMap;
uniform sampler2D refractionTex;

uniform int width;
uniform int height;

void calcSunLight(vec3 n, vec3 E, inout vec3 diffuseIntensities, inout vec3 specularIntensities) {
	vec3 l = (V * vec4(sunLight.position, 1)).xyz - fragPosition_cameraspace; 
	float d = length(l) * 0.008;
	l = normalize(l) * -1;
	// cosine of the angle between the normal and the light direction (lambert)
	float cosTheta = clamp(dot(l,n), 0, 1);
		
	float totalAttenuation =sunLight.attenuation[0] + sunLight.attenuation[1] * d + sunLight.attenuation[2]  * d * d;

	diffuseIntensities += (1.0 / totalAttenuation) * sunLight.diffColor * cosTheta;
	//diffuseIntensities += cosTheta * 100000;

	// calculate specular parts

	// Direction in which the triangle reflects the light
	vec3 R = reflect(-l,n);
	// Cosine of the angle between the Eye vector and the Reflect vector,
	float cosAlpha = clamp( dot( E,R ), 0,1 );

	specularIntensities += (1.0 / totalAttenuation) * sunLight.specColor * pow(cosAlpha, matShininess);
}

void main(){
	vec4 diffColor = vec4(0.04, 0.08, 0.14, 1.0);
	
	//vec4 diffColor = vec4(0.027, 0.52, 0.737, 1.0);


	vec3 n = normalize(normal_cameraspace); //normalize because interpolated

	// Eye vector (towards the camera)
	vec3 E = normalize(eyeDirection_cameraspace);

	vec3 diffuseIntensities = vec3(0);
	vec3 specularIntensities = vec3(0);
	for (uint i = 0; i < nbDirectionalLights; i++) {		
		vec3 l = normalize(vec3(V * vec4(directionalLights[i].direction, 0))); //normalize because interpolated
		// cosine of the angle between the normal and the light direction (lambert)
		float cosTheta = clamp(dot(l,n), 0, 1);

		//diffuseIntensities += directionalLights[i].diffColor * cosTheta;

		// calculate specular parts
		// Direction in which the triangle reflects the light
		vec3 R = reflect(-l,n);
		// Cosine of the angle between the Eye vector and the Reflect vector
		float cosAlpha = clamp( dot( E,R ), 0, 1 );

		//specularIntensities += directionalLights[i].specColor * pow(cosAlpha, matShininess * 0.03);
	}


	for (uint i = 0; i < nbPointLights; i++) {			
		vec3 l = (V * vec4(pointLights[i].position, 1)).xyz - fragPosition_cameraspace; 
		float d = length(l);
		l = normalize(l);
		// cosine of the angle between the normal and the light direction (lambert)
		float cosTheta = clamp(dot(l,n), 0, 1);
		
		float totalAttenuation = pointLights[i].attenuation[0] + pointLights[i].attenuation[1] * d + pointLights[i].attenuation[2] * d * d;

		diffuseIntensities += (1.0 / totalAttenuation) * pointLights[i].diffColor * cosTheta;

		// calculate specular parts

		// Direction in which the triangle reflects the light
		vec3 R = reflect(-l,n);
		// Cosine of the angle between the Eye vector and the Reflect vector,
		float cosAlpha = clamp( dot( E,R ), 0,1 );

		specularIntensities += (1.0 / totalAttenuation) * pointLights[i].specColor * pow(cosAlpha, matShininess);
	}

	// handle sun light
	calcSunLight(n, E, diffuseIntensities, specularIntensities);

	color = vec4(
				diffColor.xyz * matIlluminationWeights.x * 4 + 
				(diffColor.xyz * diffuseIntensities) * matIlluminationWeights.y * 30 + 
				(vec3(1) * specularIntensities) * matIlluminationWeights.z * 100, 
			1);

	// refraction 

	vec3 normalizedNormal = normalize(normal.xyz);

	vec2 refrDeviation = normalizedNormal.xz * vec2(0.08, 0.08) * normalizedNormal.y;
	vec2 refrUV = clamp(vec2(gl_FragCoord.x/width, gl_FragCoord.y/height) * 0.84 + 0.08 + refrDeviation, vec2(0.0), vec2(.99));

	// fresnel term
	float fresnel = pow(1 - clamp(dot(n, E), 0,1 ), 2.8);

	// refractionTex was already correctly lighted, so lighting is not necessary on this texture any more
	vec4 refractionColor = texture(refractionTex, refrUV);
	color = (fresnel * 0.7 * color + (1-fresnel) * refractionColor);

	//color = vec4(normalize(normal) * 0.5 + 0.5, 1);
	//color = vec4(sunLight.attenuation, 1);
}